#include <setjmp.h>

/************************************
Platform
************************************/
#define __PLAT_PC
//#define __PLAT_PS2

/************************************
Forward Declarations
************************************/
struct routineProcedure;
typedef void (*PF_RPMAIN) (routineProcedure* a_pProc); // Definition for procedure function

/************************************
User Functions
************************************/
void StartRoutine  (routineProcedure* a_pRoutine, PF_RPMAIN   a_pFunction);
void ProcessRoutine(routineProcedure* a_pRoutine);

/************************************
Constants
************************************/
static const int STACK_SIZE = 128;  // Default size of stack

/************************************
Data Structures
************************************/
struct routineProcedure 
{
    enum eR_RUNSTATE
        {
        RS_INVALID = 0,
        RS_START,
        RS_RUN
        };

    char    m_Stack[STACK_SIZE];
    int     m_StackBegin;
    int     m_StackUsed;

    int     m_PauseTime;

    jmp_buf* m_pExecSystem;
    jmp_buf  m_pExecCurrent;

    eR_RUNSTATE m_RunState;
    PF_RPMAIN   m_pFunction;

    int     m_UserParam1;  // Any parameters or structs you want passed in.
};


/************************************
Macros
************************************/
#ifdef __PLAT_PC

#define TG_SAVE_STACK(var)  __asm {mov var, ESP}  // Copy to the stack pointer to local variable
#define TG_RESTORE_STACK __asm {sub ESP, stack}   // Restore the stack with our own variable

#elif defined __PLAT_PS2

#define TG_SAVE_STACK(var)  asm ("sw   $sp,%0" : "=m" (var));
#define TG_RESTORE_STACK asm ("sub  $sp,$sp,%0" : "=g" (a_Proc.m_StackUsed));

#endif

// Pause macro  
// NOTE: If you get assert (!(a_pProc->m_StackUsed >= STACK_SIZE)); --
// You used up all your default stack space, increase the size of the stack!
#define PAUSE(time) if (setjmp(a_pProc->m_pExecCurrent) == 0) { \
                        int volatile stackEnd; \
                        TG_SAVE_STACK(stackEnd) \
                        a_pProc->m_PauseTime = time; \
                        a_pProc->m_StackUsed = a_pProc->m_StackBegin-stackEnd; \
                        assert (!(a_pProc->m_StackUsed >= STACK_SIZE)); \
                        memcpy(a_pProc->m_Stack, (void*)stackEnd, a_pProc->m_StackUsed); \
                        longjmp(*(a_pProc->m_pExecSystem), 1); }
